[Setting name="Show output window", category="General"]
bool showOutputWindow = false;

[Setting name="Enable plugin", category="General"]
bool isEnabled = false;

int outputWindowFlags = UI::WindowFlags::NoResize | UI::WindowFlags::NoCollapse | UI::WindowFlags::AlwaysAutoResize | UI::WindowFlags::NoFocusOnAppearing & UI::WindowFlags::AlwaysUseWindowPadding;

array<double> preciseTriggerTimesPool;
array<double> preciseFinishTimeCandidates;
double preciseFinishTimeFinal = 0.0;

Dev::HookInfo@ PTT   = null;
Dev::HookInfo@ SPFTC = null;
Dev::HookInfo@ EPFTC = null;
bool isSetup = false;


void Setup()
{
  if(!isSetup)
  {
#if MP41 && MANIA64
    @PTT   = Dev::Hook(Dev::BaseAddress() + 0x73D24D, 8, "PreciseTriggerTimeHook");
    @SPFTC = Dev::Hook(Dev::BaseAddress() + 0xEB449B, 2, "SavePreciseFinishTimeCandidatesHook");
    @EPFTC = Dev::Hook(Dev::BaseAddress() + 0xEC0005, 2, "EvaluatePreciseFinishTimeCandidatesHook");
    isSetup = true;
#elif TURBO
    // TBA
	print("Nanosecond functionality is not yet available for Turbo, but it is planned for the future");
#endif
    if(isSetup)
    {
      print("Enabled Nanosecond functionality");
    }
    else
    {
      print("Failed to enable Nanosecond functionality");
    }
  }
}

void Stop()
{
  if(isSetup)
  {
    Dev::Unhook(EPFTC);
    Dev::Unhook(SPFTC);
    Dev::Unhook(PTT);
	
	isSetup = false;
	
	print("Disabled Nanosecond functionality");
  }
}


void OnSettingsChanged()
{
  if(isEnabled)
  {
    Setup();
  }
  else
  {
    Stop();
  }
}


void RenderInterface()
{
  if(showOutputWindow) {
    if(UI::Begin("Nanoseconds", showOutputWindow, outputWindowFlags)) {
      UI::Text("Race time: " + FormatPreciseTime(preciseFinishTimeFinal));
    }
    UI::End();
  }
}


void PreciseTriggerTimeHook(uint64 rdx)
{
  double preciseTriggerTime = double(rdx) / 262144.0;
  preciseTriggerTimesPool.InsertLast(preciseTriggerTime);
}

void SavePreciseFinishTimeCandidatesHook()
{
  preciseFinishTimeCandidates = preciseTriggerTimesPool;
  preciseTriggerTimesPool = array<double>();
}

void EvaluatePreciseFinishTimeCandidatesHook(uint64 rcx, int32 ebx)
{
  double preciseFinishTime = 1e+300;
  for(uint i = 0; i < preciseFinishTimeCandidates.Length; i++)
  {
    if(preciseFinishTimeCandidates[i] < preciseFinishTime)
	{
	  preciseFinishTime = preciseFinishTimeCandidates[i];
	}
  }
  
  int32 timeOffset = Dev::ReadInt32(rcx + 0x58);
  
  preciseFinishTimeFinal = (preciseFinishTime - timeOffset) / 1000.0;
  
  showOutputWindow = true;
  
  print("Race Time: " + FormatPreciseTime(preciseFinishTimeFinal));
}


string FormatPreciseTime(double time)
{
  int32 truncTime = int(time * 1000);
  return Time::Format(truncTime) + Text::Format("%06d", int((time * 1000 - truncTime) * 1e+6));
}


void Main()
{
  if(isEnabled)
  {
    Setup();
  }
}

void OnDestroyed()
{
  Stop();
}